-------------------------------------------------------------------------------
--! @file       CipherCore_Control.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Control unit for Tiaoxin
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity CipherCore_Control is
    generic (
        G_MAX_LEN       : integer := SINGLE_PASS_MAX;
        G_LBS_BYTES     : integer := 5
    );
    port (
        clk             : in  std_logic;
        rst             : in  std_logic;

        --! Input
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;

        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);

        --! Datapath
        len_a           : out std_logic_vector(64               -1 downto 0);
        len_d           : out std_logic_vector(64               -1 downto 0);
        en_key          : out std_logic;
        en_state        : out std_logic;    
        sel_data        : out std_logic_vector(3                -1 downto 0);
        sel_init        : out std_logic;
        sel_decout      : out std_logic;
        sel_tag         : out std_logic;

        --! Output
        bdo_ready       : in  std_logic;
        bdo_valid       : out std_logic;
        msg_auth_done   : out std_logic
    );
end entity CipherCore_Control;

architecture behavior of CipherCore_Control is
    constant ROUND_INIT  : integer := 15;
    constant ROUND_FINAL : integer := 20;
    constant ZEROS       : std_logic_vector(128                 -1 downto 0)
                         := (others => '0');
                         
    type state_type is (
        S_START,            S_KEY_CHECK,
        S_LDKEY_1,          S_LDKEY_2,
        S_LDKEY_3,          S_LDKEY_4,
        S_WAIT_NPUB,        S_INIT,
        S_DATA,             S_LEN,
        S_FINAL,            S_TAG_OUT);
    signal state, nstate   : state_type;

    signal ctr            : std_logic_vector(5               -1 downto 0);
    signal en_ctr         : std_logic;
    signal ld_ctr         : std_logic;
    signal bdi_rdy        : std_logic;
    signal bdo_vld        : std_logic;
    signal is_decrypt     : std_logic;
    signal is_eoi         : std_logic;

    --! Note: Data represented in byte
    signal len_a_r        : std_logic_vector(G_MAX_LEN       -1 downto 0);
    signal len_d_r        : std_logic_vector(G_MAX_LEN       -1 downto 0);
    
begin
    bdi_ready <= bdi_rdy;
    bdo_valid <= bdo_vld;

    p_reg:
    process( clk )
    begin
        if rising_edge( clk ) then
            if rst = '1' then
                state         <= S_START;                
            else
                state <= nstate;

                if (ld_ctr = '1') then
                    ctr <= (others => '0');
                elsif (en_ctr = '1') then
                    ctr <= std_logic_vector(unsigned(ctr) + 1);
                end if;

                if bdi_rdy = '1' then
                    is_decrypt <= decrypt;
                    is_eoi     <= bdi_eoi;
                end if;

                if state = S_START then
                    len_a_r <= (others => '0');
                    len_d_r <= (others => '0');
                else
                    if (bdi_rdy = '1' and bdi_valid = '1') then
                        if (bdi_type(2 downto 1) = BDI_TYPE_ASS) then
                            len_a_r <= std_logic_vector(unsigned(len_a_r)
                                        + unsigned(bdi_size));
                        elsif (bdi_type(2 downto 1) = BDI_TYPE_DAT) then
                            len_d_r <= std_logic_vector(unsigned(len_d_r)
                                        + unsigned(bdi_size));
                        end if;
                    end if;
                end if;
            end if;
        end if;
    end process;

    --! Note: G_MAX_LEN < 64
    len_a <= ZEROS((64-G_MAX_LEN)-1 downto 0) & len_a_r;
    len_d <= ZEROS((64-G_MAX_LEN)-1 downto 0) & len_d_r;

    p_state:
    process( state, bdi_valid, bdi_size, bdi_type, bdi_eot, bdi_eoi, ctr,
        key_valid, key_update, bdo_ready, is_decrypt, is_eoi)
    begin
        --! Default values
        nstate          <= state;
        bdi_rdy         <= '0';
        bdo_vld         <= '0';
        key_ready       <= '0';
        msg_auth_done   <= '0';
        
        ld_ctr          <= '0';
        en_ctr          <= '0';
        en_key          <= '0';
        en_state        <= '0';
        sel_data        <= (others => '0');
        sel_init        <= '0';
        sel_decout      <= '0';
        sel_tag         <= '0';


        case state is
            when S_START =>
                nstate       <= S_KEY_CHECK;
                ld_ctr       <= '1';

            when S_KEY_CHECK =>
                if (key_update = '0' and bdi_valid = '1') then
                    nstate      <= S_WAIT_NPUB;
                elsif (key_update = '1' and key_valid = '1') then
                    nstate      <= S_LDKEY_1;
                end if;

            when S_LDKEY_1 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    nstate <= S_LDKEY_2;
                    en_key <= '1';
                end if;

            when S_LDKEY_2 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    nstate <= S_LDKEY_3;
                    en_key <= '1';
                end if;

            when S_LDKEY_3 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    nstate <= S_LDKEY_4;
                    en_key <= '1';
                end if;

            when S_LDKEY_4 =>
                key_ready <= '1';
                if (key_valid = '1') then
                    nstate <= S_WAIT_NPUB;
                    en_key <= '1';
                end if;

            when S_WAIT_NPUB =>
                bdi_rdy  <= '1';
                en_state <= '1';
                sel_init <= '1';
                if (bdi_valid = '1') then
                    nstate <= S_INIT;
                end if;

            when S_INIT =>
                en_state <= '1';
                sel_data <= "010";
                en_ctr   <= '1';
                if (unsigned(ctr) = natural(ROUND_INIT-1)) then
                    if (is_eoi = '1') then
                        nstate <= S_LEN;
                    else
                        nstate <= S_DATA;
                    end if;
                end if;

            when S_DATA =>                
                if (bdi_type(1) = '1') then
                    sel_decout <= is_decrypt;
                end if;
                if (bdi_valid = '1' 
                    and (bdo_ready = '1' or bdi_type(2 downto 1) = BDI_TYPE_ASS)) 
                then
                    en_state <= '1';
                    bdi_rdy <= '1';
                    if (bdi_type(2 downto 1) = BDI_TYPE_DAT) then
                        bdo_vld <= '1';
                        sel_data <= is_decrypt & "00";
                    elsif (bdi_type(2 downto 1) = "10") then
                        ld_ctr   <= '1';
                        en_state <= '1';
                        sel_data <= "111";                        
                    end if;
                    if (bdi_eoi = '1') then
                        nstate <= S_LEN;
                    end if;
                end if;

            when S_LEN =>
                ld_ctr   <= '1';
                en_state <= '1';
                sel_data <= "111";
                nstate <= S_FINAL;
            
            when S_FINAL =>
                en_state <= '1';
                en_ctr   <= '1';    
                sel_data <= "011";
                if (unsigned(ctr) = natural(ROUND_FINAL-1)) then
                    nstate <= S_TAG_OUT;
                end if;

            when S_TAG_OUT =>
                sel_tag    <= '1';
                if (bdi_valid = '1' and is_decrypt = '1') then
                    nstate        <= S_START;
                    bdi_rdy       <= '1';
                    msg_auth_done <= '1';
                elsif (bdo_ready = '1' and is_decrypt = '0') then
                    nstate        <= S_START;
                    bdo_vld       <= '1';
                end if;

        end case;
    end process;

end behavior;

